from django.template import Library, Node
from django.conf import settings
from bs4 import BeautifulSoup
register = Library()


import logging
log = logging.getLogger('gravy.templatetags.gravytags')


__all__ = []


class StaticOnceNode(Node):
    def __init__(self, nodelist):
        self.nodelist = nodelist

    @staticmethod
    def _squash(soup, element='script', source_tag='src'):
        srcs = set()
        for element in soup.find_all(element):
            src = element.get(source_tag)
            if src is None or not src.startswith(settings.STATIC_URL):
                continue
            if src in srcs:
                element.decompose()
                continue
            srcs.add(src)

    def render(self, context):
        output = self.nodelist.render(context)
        soup = BeautifulSoup(output, 'html.parser')
        self._squash(soup, 'script', 'src')
        self._squash(soup, 'link', 'href')
        return str(soup)


class PrettyHtmlNode(Node):
    def __init__(self, nodelist):
        self.nodelist = nodelist

    def render(self, context):
        output = self.nodelist.render(context)
        soup = BeautifulSoup(output, 'html.parser')
        return soup.prettify(formatter='html')


@register.tag
def staticonce(parser, token):
    """
    Remove duplicate static link and script (CSS and javascript) sources. The
    first reference is kept when a duplicate is found.
    """
    nodelist = parser.parse(('endstaticonce',))
    parser.delete_first_token()
    return StaticOnceNode(nodelist)


@register.tag
def prettyhtml(parser, token):
    """
    Make the html in this tag pretty.
    """
    nodelist = parser.parse(('endprettyhtml',))
    parser.delete_first_token()
    return PrettyHtmlNode(nodelist)



from django.forms.widgets import *
from django.utils.safestring import mark_safe
from django.utils.html import escape
from django.utils import six
from django.core.files import File
from django.http import QueryDict
from ..utils import datetime
from collections import OrderedDict
import re


import logging
log = logging.getLogger('gravy.forms.widgets')


__all__ = [
    'NamedMultiWidget', 'RepeatNamedMultiWidget', 'SeparatedWidgetMixin',
    'SeparatedSelect', 'SeparatedTextInput', 'SeparatedTextarea',
    'SerializedDateTimeInput', 'MultipleFileInput',
]
# proxy django.forms.widgets
import django.forms.widgets
__all__.extend(django.forms.widgets.__all__)


# XXX: don't bother inheriting here? Use lower level mixins?
# XXX: implement __deepcopy__
class NamedMultiWidget(MultiWidget):
    widgets = ()
    subwidget_name_format = '{0}.{1}'

    class Media:
        css = {
            'all': ('css/jquery.namedMultiWidget.css',)
        }

    def __init__(self, widgets=None, attrs=None, labels=None, help_texts=None, initial=None):
        self.named_widgets = OrderedDict(widgets or self.widgets)
        self._labels = labels or {}
        self._help_texts = help_texts or {}
        self._initial = initial or {}
        widgets = self.named_widgets.values()
        super(NamedMultiWidget, self).__init__(widgets, attrs=attrs)

    def render(self, name, value, attrs=None):
        if self.is_localized:
            for widget in self.widgets:
                widget.is_localized = self.is_localized
        if not isinstance(value, list):
            value = self.decompress(value)
        output = []
        final_attrs = self.build_attrs(attrs)
        id_ = final_attrs.get('id', None)
        for i, subwidget in enumerate(self.named_widgets.items()):
            widget_name, widget = subwidget
            widget_attrs = dict(final_attrs)
            widget_attrs.pop('id', None)
            try:
                widget_value = value[i]
            except IndexError:
                widget_value = None
            widget_name = self.subwidget_name_format.format(name, widget_name)
            output.append(widget.render(widget_name, widget_value, widget_attrs))
        return mark_safe(self.format_output(output, attrs=final_attrs))

    def id_for_label(self, id_):
        return id_

    def value_from_datadict(self, data, files, name):
        value = []
        for widget_name, widget in six.iteritems(self.named_widgets):
            widget_name = self.subwidget_name_format.format(name, widget_name)
            value.append(widget.value_from_datadict(data, files, widget_name))
        return value

    def format_output(self, rendered_widgets, attrs=None):
        id_ = attrs.get('id')
        id_ = ' id="{0}"'.format(id_) if id_ else ''
        ul = '<ul{0} class="named-multi-widget">{2}{1}</ul>'
        li = '<li>{0}{1}</li>'
        title = ' title="{0}"'
        label = '<label{1}>{0}</label>'
        rendered = []
        hidden = []
        for i, subwidget in enumerate(self.named_widgets.items()):
            widget_name, widget = subwidget
            if widget.is_hidden:
                hidden.append(rendered_widgets[i])
            else:
                t = self._help_texts.get(widget_name)
                t = title.format(t) if t else ''
                l = self._labels.get(widget_name)
                l = label.format(l, t) if l else ''
                rendered.append(li.format(l, rendered_widgets[i]))
        return ul.format(id_, ''.join(rendered), ''.join(hidden))

    def decompress(self, value):
        if value:
            return [value.get(name, None) for name in self.named_widgets.keys()]
        return [None]*len(self.named_widgets)


class RepeatNamedMultiWidget(NamedMultiWidget):

    widget_class = 'repeat-named-multi-widget'
    toggle = 'repeatNamedMultiWidget'

    class Media:
        css = {
            'all': ('css/jquery.repeatNamedMultiWidget.css',)
        }
        js = ('js/jquery.repeatNamedMultiWidget.js',)

    _id_fmt = ' id="{id}"'
    _ul_fmt = '<ul{id} class="{classname}" data-toggle="{toggle}" data-repeat-template="{template}">{subwidgets}<span class="{classname}-add"></span></ul>'
    _li_fmt = '<li><span class="{classname}-remove"></span>{subwidget}</li>'
    _nin_re = re.compile(r'(?P<name>[^\[]+)\[(?P<index>\d+)\](?P<next>.*)')

    def _get_lists(self, d):
        if hasattr(d, 'lists'):
            return d.lists()
        return [(n, [v]) for n, v in d.items()]

    def _get_subquerydicts(self, name, querydict):
        subquerydicts = {}
        querylists = self._get_lists(querydict)
        for n, valuelist in querylists:
            match = self._nin_re.match(n)
            if match is None or match.group('name') != name:
                continue
            index = int(match.group('index'))
            subquerydicts.setdefault(index, QueryDict(mutable=True))
            subquerydicts[index].setlist(name + match.group('next'), valuelist)
        return subquerydicts

    def value_from_datadict(self, data, files, name):
        dsqd = self._get_subquerydicts(name, data)
        fsqd = self._get_subquerydicts(name, files)
        dfsqd = {}
        for i in dsqd:
            dfsqd[i] = (dsqd[i], fsqd.pop(i, {}))
        for i in fsqd:
            dfsqd[i] = ({}, fsqd[i])
        dfsql = [dfsqd[i] for i in sorted(dfsqd)]
        value = []
        for data, files in dfsql:
            value.append(super(RepeatNamedMultiWidget, self).value_from_datadict(data, files, name))
        return value

    def render(self, name, values, attrs=None):
        name += '[]'
        final_attrs = self.build_attrs(attrs)
        widget_attrs = dict(final_attrs)
        widget_attrs.pop('id', None)
        template_value = self._initial
        if not isinstance(values, list):
            template_value, values = values, []
        template = super(RepeatNamedMultiWidget, self).render(name, template_value, widget_attrs)
        rendered_widgets = []
        for i, value in enumerate(values):
            rendered_widgets.append(super(RepeatNamedMultiWidget, self).render(name, value, widget_attrs))
        return mark_safe(self.format_output_repeat(template, rendered_widgets, attrs=final_attrs))

    def format_output_repeat(self, template, rendered_widgets, attrs=None):
        if self.is_hidden:
            return ''.join(rendered_widgets)
        id_ = attrs.get('id')
        id_ = self._id_fmt.format(id=id_) if id_ else ''
        template = escape(self._li_fmt.format(classname=self.widget_class, subwidget=template))
        rendered = []
        for i, rendered_widget in enumerate(rendered_widgets):
            rendered.append(self._li_fmt.format(classname=self.widget_class, subwidget=rendered_widget))
        return self._ul_fmt.format(classname=self.widget_class,
            toggle=self.toggle, id=id_, template=template,
            subwidgets=''.join(rendered)
        )


class SeparatedWidgetMixin(object):

    def __init__(self, attrs=None, token=',', cleanup=True, **kwargs):
        self.token = token
        self.cleanup = cleanup
        super(SeparatedWidgetMixin, self).__init__(attrs=attrs, **kwargs)

    def render(self, name, value, attrs=None, **kwargs):
        if isinstance(value, list):
            value = self.token.join([unicode(v) for v in value])
        return super(SeparatedWidgetMixin, self).render(name, value, attrs=attrs, **kwargs)

    def value_from_datadict(self, data, files, name):
        value = super(SeparatedWidgetMixin, self).value_from_datadict(data, files, name)
        if not isinstance(value, six.string_types):
            return value
        values = value.split(self.token)
        if self.cleanup:
            values = filter(bool, [v.strip() for v in values])
        return values


class SeparatedSelect(SeparatedWidgetMixin, Select):

    def _collapse_choices(self, choices):
        for choice in choices:
            if isinstance(choice[0], list):
                choice = (self.token.join([unicode(c) for c in choice[0]]), choice[1])
            yield choice

    def render(self, name, value, attrs=None, choices=()):
        choices = list(self._collapse_choices(choices))
        self.choices = list(self._collapse_choices(self.choices))
        return super(SeparatedSelect, self).render(name, value, attrs=attrs, choices=choices)


class SeparatedTextInput(SeparatedWidgetMixin, TextInput):
    pass


class SeparatedTextarea(SeparatedWidgetMixin, Textarea):
    pass


class SerializedDateTimeInput(DateTimeInput):

    def _format_value(self, value):
        if isinstance(value, six.integer_types):
            value = datetime(value)
        return super(SerializedDateTimeInput, self)._format_value(value)


class MultipleFileInput(FileInput):

    def render(self, name, value, attrs=None):
        attrs['multiple'] = 'multiple'
        return super(MultipleFileInput, self).render(name, value, attrs)

    def value_from_datadict(self, data, files, name):
        if hasattr(files, 'getlist'):
            return files.getlist(name)
        else:
            value = files.get(name)
            if isinstance(value, list):
                return value
            else:
                return [value]



from django.forms.models import ModelMultipleChoiceField
from django.forms.widgets import *
from django.forms.fields import *
from django.conf import settings
from django.db.models.fields import TextField
from django.utils import formats
from .fields import *
from .widgets import *


__all__ = [
    'BootstrapTextInput', 'BootstrapTextarea', 'BootstrapNumberInput',
    'BootstrapSelect', 'BootstrapSeparatedSelect', 'BootstrapDateTimePicker',
    'BootstrapSerializedDateTimePicker', 'BootstrapFileInput',
    'BootstrapCheckboxToggle', 'BootstrapCombobox', 'BootstrapDateRangePicker',
    'BootstrapSeparatedTextarea', 'BootstrapMultipleFileInput',
]


class BootstrapMixin(object):
    classes = None
    toggle = None

    def __init__(self, attrs=None, classes=None, toggle=None, *args, **kwargs):
        if attrs is None:
            attrs = {}
        classes = self.classes if classes is None else classes
        if classes:
            attrs['class'] = ' '.join(classes)
        toggle = self.toggle if toggle is None else classes
        if toggle:
            attrs['data-toggle'] = toggle
        super(BootstrapMixin, self).__init__(attrs=attrs, *args, **kwargs)


class FormControlMixin(object):
    classes = ('form-control',)


class OptionalMixin(object):

    def render(self, name, value, attrs=None):
        if not self.is_required:
            if attrs is None:
                attrs = {}
            attrs['placeholder'] = 'Optional'
        return super(OptionalMixin, self).render(name, value, attrs=attrs)


class BootstrapTextInput(OptionalMixin, FormControlMixin,  BootstrapMixin, TextInput):
    pass
    

class BootstrapTextarea(OptionalMixin, FormControlMixin, BootstrapMixin, Textarea):
    pass


class BootstrapNumberInput(FormControlMixin, BootstrapMixin, NumberInput):
    pass


class BootstrapSelect(FormControlMixin, BootstrapMixin, Select):
    pass


class BootstrapSelectMultiple(FormControlMixin, BootstrapMixin, SelectMultiple):
    pass


class BootstrapSeparatedSelect(FormControlMixin, BootstrapMixin, SeparatedSelect):
    pass


class BootstrapSeparatedTextarea(OptionalMixin, FormControlMixin, BootstrapMixin, SeparatedTextarea):
    pass


class DateTimeMixin(object):
    toggle = 'datetimePicker'

    class Media:
        css = {'all': ('css/bootstrap-datetimepicker.css',)}
        js = (
            'js/moment.min.js',
            'js/bootstrap-datetimepicker.min.js',
            'js/bootstrap.datetimePicker.js',
        )

    def __init__(self, attrs=None, format=None):
        attrs = attrs or {}
        attrs['data-widget-positioning'] = '{"horizontal": "right", "vertical": "bottom"}'
        attrs['data-format'] = self._moment_format(format or formats.get_format(self.format_key)[0])
        super(DateTimeMixin, self).__init__(attrs=attrs, format=format)

    _format_map = (
        ('DDD', r'%j'),
        ('DD', r'%d'),
        ('MMMM', r'%B'),
        ('MMM', r'%b'),
        ('MM', r'%m'),
        ('YYYY', r'%Y'),
        ('YY', r'%y'),
        ('HH', r'%H'),
        ('hh', r'%I'),
        ('mm', r'%M'),
        ('ss', r'%S'),
        ('a', r'%p'),
        ('ZZ', r'%z'),
    )

    @classmethod
    def _moment_format(cls, fmt):
       for m, d in cls._format_map:
            fmt = fmt.replace(d, m)
       return fmt


class BootstrapDateTimePicker(OptionalMixin, FormControlMixin, DateTimeMixin, BootstrapMixin, DateTimeInput):
    pass


class BootstrapSerializedDateTimePicker(OptionalMixin, FormControlMixin, DateTimeMixin, BootstrapMixin, SerializedDateTimeInput):
    pass


class BootstrapFileInput(OptionalMixin, FormControlMixin, BootstrapMixin, FileInput):
    toggle = 'fileInput'

    class Media:
        css = {'all': ('css/bootstrap.fileInput.css',)}
        js = ('js/bootstrap.fileInput.js',)


class BootstrapCheckboxToggle(BootstrapMixin, CheckboxInput):
    toggle = 'checkboxToggle'

    class Media:
        css = {'all': ('css/bootstrap.checkboxToggle.css',)}
        js = ('js/bootstrap.checkboxToggle.js',)


class BootstrapCombobox(FormControlMixin, BootstrapMixin, Select):
    toggle = 'combobox'

    class Media:
        css = {'all': ('css/bootstrap-combobox.css',)}
        js = (
            'js/bootstrap-combobox.js',
            'js/bootstrap.combobox.js'
        )


class BootstrapDateRangePicker(NamedMultiWidget):
    date_ranges = (
        (1800, '30 min'),
        (3600, '1 hr'),
        (5400, '1.5 hr'),
        (7200, '2 hr'),
    )

    class Media:
        css = {'all': ('css/bootstrap.dateRangePicker.css',)}
        js = ('js/bootstrap.dateRangePicker.js',)

    def __init__(self, date_ranges=None, *args, **kwargs):
        if date_ranges is not None:
            self.date_ranges = date_ranges
        super(BootstrapDateRangePicker, self).__init__(*args, **kwargs)

    def get_date_ranges(self):
        return self.date_ranges

    def format_output(self, rendered_widgets, attrs=None):
        div = '<div data-toggle="dateRangePicker">{0}<div class="btn-group">{1}</div></div>'
        btn = '<button class="btn btn-default" data-value="{0}">{1}</button>'
        btn_group = ''.join([btn.format(v, n) for v, n in self.get_date_ranges()])
        output = super(BootstrapDateRangePicker, self).format_output(rendered_widgets, attrs=attrs)
        return div.format(output, btn_group)


class BootstrapMultipleFileInput(OptionalMixin, FormControlMixin, BootstrapMixin, MultipleFileInput):
    toggle = 'fileInput'

    class Media:
        css = {'all': ('css/bootstrap.fileInput.css',)}
        js = ('js/bootstrap.fileInput.js',)


# theme
# XXX: it turns out this was a bad idea as it also modifies the admin
# interface. Make BootstrapFormMixin and BootstrapModelFormMixin instead.
_Textarea_render_original = Textarea.render
def _Textarea_render(self, name, value, attrs=None):
    if attrs is None:
        attrs = {}
    if 'class' in attrs:
        attrs['class'] += ' form-control'
    else:
        attrs['class'] = 'form-control'
    return _Textarea_render_original(self, name, value, attrs=attrs) 


def apply_theme():
    CharField.widget = BootstrapTextInput
    BooleanField.widget = BootstrapCheckboxToggle
    IntegerField.widget = BootstrapNumberInput
    ChoiceField.widget = BootstrapSelect
    FileField.widget = BootstrapFileInput
    MultipleFileField.widget = BootstrapMultipleFileInput
    DateTimeField.widget = BootstrapDateTimePicker
    SerializedDateTimeField.widget = BootstrapSerializedDateTimePicker
    ModelMultipleChoiceField.widget = BootstrapSelectMultiple
    SeparatedChoiceField.widget = BootstrapSeparatedSelect
    Textarea.render = _Textarea_render


if getattr(settings, 'FORMS_THEME', None) == 'bootstrap':
    apply_theme()



from django.conf.urls import url
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from .namespace import NamespaceHandlerRegistry


import logging
log = logging.getLogger('gravy.websocket')


__all__ = ['urls',]


@csrf_exempt
def ws(request, namespace=None):
    handler_class = NamespaceHandlerRegistry.get(namespace)
    handler_class(request).run()
    return HttpResponse()


urls = [url(r'^(?P<namespace>\w+)$', ws)]



from django.core.files.storage import FileSystemStorage
import hashlib
import base64


__all__ = ['HashedFileSystemStorage', 'B64HashedFileSystemStorage',]


class HashedFileSystemStorage(FileSystemStorage):
    """
    Use the hexdigest of the hashed file content as the filename. If the
    filename already exists assume it has the same content and don't bother
    writing it to disk.
    """
    algorithm = 'sha1'

    def __init__(self, algorithm=None, *args, **kwargs):
        if algorithm is not None:
            self.algorithm = algorithm
        super(HashedFileSystemStorage, self).__init__(*args, **kwargs)

    def get_available_name(self, name):
        return name

    def hash_to_name(self, hash):
        return hash.hexdigest()

    def _save(self, name, content):
        h = hashlib.new(self.algorithm)
        for chunk in content.chunks():
            h.update(chunk)
        name = self.hash_to_name(h)
        if self.exists(name):
            return name
        return super(HashedFileSystemStorage, self)._save(name, content)


class B64HashedFileSystemStorage(HashedFileSystemStorage):
    """
    Use the base64 encoded digest of the hashed file content as the filename.
    The base64 padding characters are removed from the filename.
    """

    def hash_to_name(self, hash):
        return base64.b64encode(hash.digest()).rstrip('=')



from django.utils.six.moves.urllib.parse import unquote
from django.utils import timezone, six
from django.template import Template, Context
from django.conf import settings
from collections import OrderedDict
from datetime import datetime as _datetime
import magic
import mimetypes
import calendar
import weakref
import time
import os
import re


import logging
log = logging.getLogger('gravy.utils')


# TODO: check the datetime functions (do they do what is expected?)
def epoch(dt):
    """
    Convert a datetime object to unix epoch timestamp.
    """
    return int(time.mktime(dt.timetuple()))


def datetime(e):
    """
    Convert epoch datetime to localtime timezone aware datetime.
    """
    return timezone.make_aware(_datetime.fromtimestamp(e))


def epochutc(dt):
    """
    Convert a datetime object to unix utc epoch timestamp.
    """
    return int(calendar.timegm(dt.utctimetuple()))


def datetimeutc(e):
    """
    Convert epoch datetime to utc timezone aware datetime.
    """
    return timezone.make_aware(_datetime.utcfromtimestamp(e), timezone.utc)


def seconds_to_units(seconds):
    """
    Converts number of seconds as a int into
    weeks, days, hours, minutes, seconds delta
    """
    minutes, seconds = divmod(seconds, 60)
    if not minutes:
        return (seconds, 'second')
    hours, minutes = divmod(minutes, 60)
    if not hours:
        return (minutes, 'minute')
    days, hours = divmod(hours, 24)
    if not days:
        return (hours, 'hour')
    weeks, days = divmod(days, 7)
    if not weeks:
        return (days, 'day')
    return (weeks, 'week')


def safepath(path):
    """
    Take an untrusted path (possibly urlencoded) return a path that is safe
    from directory traversal. Taken from django.views.static.
    """
    path = os.path.normpath(unquote(path))
    path = path.lstrip('/')
    newpath = ''
    for part in path.split('/'):
        if not part:
            # Strip empty path components.
            continue
        drive, part = os.path.splitdrive(part)
        head, part = os.path.split(part)
        if part in (os.curdir, os.pardir):
            # Strip '.' and '..' in path.
            continue
        newpath = os.path.join(newpath, part).replace('\\', '/')
    return newpath


# we only want to do this once
_templify_template = Template('{{ obj }}')


def templify(obj):
    """
    Converts the object to a string as it would occur in a template if it
    were rendered inline ie. {{ variable }}. This is especially useful
    for datetime objects.
    """
    return _templify_template.render(Context({'obj': obj}))


def safe_meta(txt, tiny=True):
    """
    Remove excess whitespace and newlines. This allows us to use double
    newlines as a delimiter between metadata fields.
    """
    if tiny:
        return re.sub(r'\s+', ' ', txt)
    return re.sub(r'\n+', '\n',
        re.sub(r'(?m)^\s*$', '',
            re.sub(r'[ \t\r]+', ' ', txt)
        )
    )


def dict_meta(d):
    """
    Convert a dictionary to metadata.
    """
    return '\n'.join(
        [': '.join(map(str, [k, v])) for k, v in six.iteritems(d)]
    )


_hexdump_map = ''.join([(len(repr(chr(x)))==3) and chr(x) or '.' for x in range(256)])


def hexdump(src, length=16):
    result=[]
    for i in xrange(0, len(src), length):
       s = src[i:i+length]
       hexa = ' '.join(["%02X"%ord(x) for x in s])
       printable = s.translate(_hexdump_map)
       result.append("%04X   %-*s   %s\n" % (i, length*3, hexa, printable))
    return ''.join(result)


def magic_mimetype(data='', filename=None):
    """
    Use libmagic to determine the mimetype of the given data (file content).
    """
    if filename is None:
        return magic.from_buffer(data, mime=True)
    return magic.from_file(filename, mime=True)


def magic_type(data='', filename=None):
    """
    Use libmagic to determine the mimetype of the given data (file content).
    """
    if filename is None:
        return magic.from_buffer(data)
    return magic.from_file(filename)


# use ours by preference
mimetypes.knownfiles = getattr(settings, 'MIMETYPES', mimetypes.knownfiles)
mimetypes.types_map = {}
mimetypes.init()


def magic_extension(mimetype):
    """
    Use the mimetype to determine the file extension of the given data (file
    content).
    """
    return mimetypes.guess_extension(mimetype)


class _DoesNotExist(object):
    def __str__(self):
        return ''
    def __unicode_(self):
        return u''
    def __bool__(self):
        return False
    def __nonzero__(self):
        return False
    def __int__(self):
        return 0


DoesNotExist = _DoesNotExist()



from django.utils.decorators import method_decorator
from django.utils import timezone
from django.shortcuts import get_object_or_404
from django.contrib.messages.views import SuccessMessageMixin
from django.views.generic import UpdateView, CreateView
from django.forms.models import modelform_factory
from django.http import JsonResponse, HttpResponse, Http404
from django.contrib.auth.decorators import login_required
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import RedirectView, View
from django.core.exceptions import PermissionDenied
from datetime import timedelta, datetime
from calendar import timegm


import logging
log = logging.getLogger('gravy.views')


# Create your views here.
class LoginRequiredMixin(object):
    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super(LoginRequiredMixin, self).dispatch(*args, **kwargs)


class NeverCacheMixin(object):
    @method_decorator(never_cache)
    def dispatch(self, *args, **kwargs):
        return super(NeverCacheMixin, self).dispatch(*args, **kwargs)


class CsrfExemptMixin(object):
    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super(CsrfExemptMixin, self).dispatch(*args, **kwargs)


class SuperuserRequiredMixin(object):
    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_superuser:
            raise PermissionDenied
        return super(SuperuserRequiredMixin, self).dispatch(request, *args, **kwargs)


class JsonResponseMixin(object):
    def render_to_response(self, context, **kwargs):
        return JsonResponse(self.get_data(context), **kwargs)
    def get_data(self, context):
        return context


class BetterRedirectView(RedirectView):
    permanent = True
    extra_kwargs = {}
    def get_redirect_url(self, *args, **kwargs):
        kwargs.update(self.extra_kwargs)
        return super(BetterRedirectView, self).get_redirect_url(*args, **kwargs)


class NextUrlMixin(object):
    def get_redirect_url(self, *args, **kwargs):
        return self.request.GET.get('next', '/')


class LastPageBase(object):
    lastpage_namespace = 'lastpage'
    def get_lastpage_key(self, key):
        return ':'.join([
            self.lastpage_namespace,
            key
        ])
        

class LastPageMixin(LastPageBase):
    def get_lastpage_keys(self):
        for key in self.lastpage_keys:
            yield self.get_lastpage_key(key)
    def dispatch(self, request, *args, **kwargs):
        for key in self.get_lastpage_keys():
            request.session[key] = request.path
        return super(LastPageMixin, self).dispatch(request, *args, **kwargs)


class LastPageRedirectView(LastPageBase, BetterRedirectView):
    permanent = False
    lastpage_key = None
    def get_redirect_url(self, *args, **kwargs):
        if self.lastpage_key is not None:
            key = self.get_lastpage_key(self.lastpage_key)
            self.url = self.request.session.get(key)
        return super(LastPageRedirectView, self).get_redirect_url(*args, **kwargs)


class RelatedMixin(object):
    related = ()
    def get_queryset(self):
        qs = super(RelatedMixin, self).get_queryset()
        if self.related and self.request.method in ('GET',):
            qs = qs.select_related(*self.related)
        return qs


class DeferMixin(object):
    """
    Add deferred fields to queryset.
    Note: If you use defer on a DetailView (or any view that uses
    SingleObjectTemplateResponseMixin) the template name will be mangled as a
    proxy model is generated when defer is used.
    Note: This issue #24689 has been patched in django master.
    """
    defer = ()
    def get_queryset(self):
        qs = super(DeferMixin, self).get_queryset()
        if self.defer and self.request.method in ('GET',):
            qs = qs.defer(*self.defer)
        return qs


class AutoSuccessMessageMixin(SuccessMessageMixin):
    def get_success_message(self, cleaned_data):
        if not self.success_message:
            name = self.object.__class__.__name__
            if isinstance(self, CreateView):
                self.success_message = u'%s created successfully.' % name
            elif isinstance(self, UpdateView):
                self.success_message = u'%s updated successfully.' % name
        return super(AutoSuccessMessageMixin, self).get_success_message(cleaned_data)


class ModelFormWidgetMixin(object):
    def get_form_class(self):
        if self.form_class:
            return self.form_class
        else:
            if self.model is not None:
                model = self.model
            elif hasattr(self, 'object') and self.object is not None:
                model = self.object.__class__
            else:
                model = self.get_queryset().model
            return modelform_factory(model, fields=self.fields, widgets=self.widgets)


class BulkOperationMixin(object):
    pk_param = 'pk'

    def get_pk_list(self):
        if self.request.method == 'GET':
            return self.request.GET.getlist(self.pk_param)
        elif self.request.method == 'POST':
            return self.request.POST.getlist(self.pk_param)
        return []

    def get_queryset(self):
        pk_list = self.get_pk_list()
        return super(BulkOperationMixin, self).get_queryset().filter(pk__in=pk_list)

    def get_object_list(self):
        pk_list = self.get_pk_list()
        return self.model.objects.filter(pk__in=pk_list)


class ForeignModelMixin(object):
    foreign_context_name = 'foreign'
    foreign_force_query = False
    foreign_required = False

    def get_context_data(self, **kwargs):
        """
        Try to find a model instance to add to the context. Attempts to locate
        an instance efficiently only performing additional queries as a last
        resort.
        """
        context = super(ForeignModelMixin, self).get_context_data(**kwargs)

        # already know the instance
        if self.foreign_object is not None:
            context[self.foreign_context_name] = self.foreign_object
            return context

        # try to find an object from which to grab an instance
        obj = None
        if 'object_list' in context:
            # force evaluation (prevents double query when getting first object)
            obj_list = list(context['object_list'])
            if obj_list:
                obj = obj_list[0]
            context['object_list'] = obj_list
        elif 'object' in context:
            obj = context['object']

        # get the instance from the object when possible
        if obj is not None:
            if isinstance(obj, self.foreign_model):
                self.foreign_object = obj
            elif hasattr(obj, self.foreign_field_name):
                self.foreign_object = getattr(obj, self.foreign_field_name)
            else:
                self.foreign_object = obj[self.foreign_field_name]

        # get the instance from the provided instance pk
        elif self.foreign_object_pk is not None:
            self.foreign_object = get_object_or_404(self.foreign_model, pk=self.foreign_object_pk)

        if self.foreign_object is not None:
            context[self.foreign_context_name] = self.foreign_object

        return context

    def get_form_kwargs(self):
        """
        When a form is posted set the appropriate field to an instance of the
        model when possible.
        """
        kwargs = super(ForeignModelMixin, self).get_form_kwargs()
        if self.foreign_object_pk is not None and 'instance' in kwargs:
            if kwargs['instance'] is None:
                kwargs['instance'] = self.model()
            setattr(kwargs['instance'], self.foreign_field_name + '_id', self.foreign_object_pk)
        return kwargs

    def get_queryset(self):
        """
        Add the model instance to the queryset and filter to instance if given.
        """
        queryset = super(ForeignModelMixin, self).get_queryset()
        if issubclass(queryset.model, self.foreign_model):
            return queryset
        queryset = queryset.select_related(self.foreign_field_name)
        if self.foreign_object_pk is not None:
            filter_kwargs = {self.foreign_field_name + '_id': self.foreign_object_pk}
            queryset = queryset.filter(**filter_kwargs)
        return queryset

    def dispatch(self, request, *args, **kwargs):
        """
        Retreive a model instance pk from get params if it exists. If forced
        also load the instance object from the database.
        """
        self.foreign_object_pk = self.request.GET.get(self.foreign_field_name)
        if self.foreign_required and self.foreign_object_pk is None:
            raise Http404
        if self.foreign_object_pk is not None and self.foreign_force_query:
            self.foreign_object = get_object_or_404(self.foreign_model, pk=self.foreign_object_pk)
        else:
            self.foreign_object = None
        return super(ForeignModelMixin, self).dispatch(request, *args, **kwargs)


class FlotMixin(JsonResponseMixin):

    SECOND = 1
    MINUTE = SECOND * 60
    HOUR = MINUTE * 60
    DAY = HOUR * 24
    WEEK = DAY * 7

    # helpers
    @staticmethod
    def _datetime(je):
        return timezone.make_aware(
            datetime.utcfromtimestamp(je / 1000), timezone.utc
        )

    @staticmethod
    def _flotutc(dt):
        return timegm(dt.utctimetuple()) * 1000

    @staticmethod
    def _flotutc_sql(expr):
        return 'CAST(EXTRACT(EPOCH FROM %s AT TIME ZONE \'UTC\')*1000 AS BIGINT)' % expr
    
    @classmethod
    def _flotutc_datetime_trunc_sql(cls, lookup_type, field_name):
        return cls._flotutc_sql(
            "DATE_TRUNC('%s', %s AT TIME ZONE \'UTC\')" % (lookup_type, field_name)
        )

    @classmethod
    def _flotwidth(cls, unit):
        ms = 1000
        return {
            'second': cls.SECOND * ms,
            'minute': cls.MINUTE * ms,
            'hour': cls.HOUR * ms,
            'day': cls.DAY * ms,
            'week': cls.WEEK * ms,
        }[unit]

    def get_plot(self, context):
        return {
            'sequence': self.request.GET.get('sequence'),
            'options': {
                'colors': ['#09d809'],
                'series': {
                    'shadowSize': 0,
                    'lines': {'lineWidth': 1},
                    'points': {'lineWidth': 1, 'radius': 2, 'fill': True},
                    'bars': {'lineWidth': 1, 'align': 'center'},
                },
                'grid': {
                    'borderColor': '#eee',
                    'borderWidth': {'top': 0, 'left': 0, 'right': 0, 'bottom': 1},
                    'hoverable': True,
                },
                'zoom': {
                    'interactive': True
                },
                'pan': {
                    'interactive': True
                },
                'xaxis': {
                    'tickLength': 0,
                },
                'yaxis': {
                    'min': 0,
                    'panRange': False,
                    'zoomRange': False,
                    'tickDecimals': 0,
                    'labelWidth': 30,
                    'reserveSpace': True,
                },
            }
        }

    def get_data(self, context):
        return self.get_plot(context)


# TODO: this is a work in progress
class FlotTimeMixin(FlotMixin):
    
    MIN_POINTS = 8

    plot_duration = timedelta(hours=24)
    plot_offsetx = timedelta(hours=1)

    def apply_queryset_limits(self, queryset):

        # extend limits
        delta = self.end - self.begin
        begin = self.begin - delta
        end = self.end + delta

        field = self.plot_fieldx

        # include extra datapoint before limit
        kwargs = {field + '__lt': begin}
        self.before = queryset.values_list(field, flat=True).filter(**kwargs).first()
        if self.before is not None:
            kwargs = {field + '__gte': self.before}
            queryset = queryset.filter(**kwargs)

        # include extra datapoint after limit
        kwargs = {field + '__gt': end}
        self.after = queryset.values_list(field, flat=True).filter(**kwargs).last()
        if self.after is not None:
            kwargs = {field + '__lte': self.after}
            queryset = queryset.filter(**kwargs)

        return queryset

    def get_series(self, context):
        return []

    def get_plot(self, context):
        plot = super(FlotTimeMixin, self).get_plot(context)
        plot['series'] = self.get_series(context)
        plot['options']['xaxis']['min'] = self._flotutc(self.begin)
        plot['options']['xaxis']['max'] = self._flotutc(self.end + self.offsetx)
        plot['options']['xaxis']['mode'] = 'time'
        plot['options']['xaxis']['timezone'] = 'browser'
        plot['options']['xaxis']['zoomRange'] = [1 * 60 * 1000, 4 * 7 * 24 * 60 * 60 * 1000]
        return plot

    def get_unit(self, duration):
        for unit in ('week', 'day', 'hour', 'minute'):
            if duration / self._flotwidth(unit) >= self.MIN_POINTS:
                return unit
        return 'second'

    def get_limits(self):
        # request args
        r = self.request
        self.xmin = int(r.GET['xmin']) if 'xmin' in r.GET else None
        self.xmax = int(r.GET['xmax']) if 'xmax' in r.GET else None
        self.ymin = int(r.GET['ymin']) if 'ymin' in r.GET else None
        self.ymax = int(r.GET['ymax']) if 'ymax' in r.GET else None

        # useful extras
        self.now = timezone.now()
        self.end = self.now if self.xmax is None else self._datetime(self.xmax)
        self.begin = self.end - self.plot_duration if self.xmin is None else self._datetime(self.xmin)
        self.unit = self.get_unit((self.end-self.begin).total_seconds()*1000)
        self.offsetx = self.plot_offsetx if self.xmax is None else timedelta(0)

    def get(self, request, *args, **kwargs):
        self.get_limits()
        return super(FlotTimeMixin, self).get(request, *args, **kwargs)


class FlotPieMixin(FlotMixin):

    def get_plot(self, context):
        plot = super(FlotPieMixin, self).get_plot(context)
        plot['data'] = self.get_series(context)
        plot['options'] = {
            'series': {
                'pie': {
                    'show': True,
                    'radius': 1,
                    'label': {'show': True, 'radius': 0.6},
                    'stroke': {'width': 3},
                }
            },
            'legend': {'show': False}
        }
        return plot


class NullView(View):
    def dispatch(self, request, *args, **kwargs):
        return HttpResponse(status=204)



from django_redis import get_redis_connection
from .helpers import serialize, deserialize
import six
import gevent
import fnmatch
import os


import logging
log = logging.getLogger('gravy.websocket.redis')


__all__ = ['RedisPubsubRegistry', 'RedisMixin', 'RedisRoomsMixin',]


_redis_conn = get_redis_connection('default')


class _RedisPubsubRegistry(object):

    def __init__(self, conn):
        self._registry = {}
        self.conn = conn
        self.sub_channel = 'redispubsubregistry:sub:%d:%d' % (os.getpid(), hash(self))
        self.unsub_channel = 'redispubsubregistry:unsub:%d:%d' % (os.getpid(), hash(self))
        self.task = gevent.spawn(self.listener)

    def listener(self):
        pubsub = self.conn.pubsub()
        pubsub.subscribe(self.sub_channel)
        pubsub.subscribe(self.unsub_channel)
        for msg in pubsub.listen():
            # ignore our own psubscribes
            if msg['type'] == 'psubscribe':
                continue
            # subscribe to channel
            if msg['channel'] == self.sub_channel:
                pubsub.psubscribe(msg['data'])
                continue
            # unsubscribe from channel
            if msg['channel'] == self.unsub_channel:
                pubsub.punsubscribe(msg['data'])
                continue
            # emit messages to namespace handlers
            for channel, handlers in six.iteritems(self._registry):
                if not fnmatch.fnmatch(msg['channel'], channel):
                    continue
                for handler in handlers:
                    handler.check_and_send(msg)

    def publish(self, channel, msg):
        """
        Convienience method for publishing.
        """
        self.conn.publish(channel, msg)

    def register(self, channel, handler):
        if channel not in self._registry:
            self._registry[channel] = []
            self.conn.publish(self.sub_channel, channel)
        if handler not in self._registry[channel]:
            self._registry[channel].append(handler)
        log.debug('%s has %d sockets' % (channel, len(self._registry[channel])))

    def unregister(self, channel, handler):
        if channel in self._registry:
            if handler in self._registry[channel]:
                self._registry[channel].remove(handler)
                if not self._registry[channel]:
                    self.conn.publish(self.unsub_channel, channel)
                    del self._registry[channel]


RedisPubsubRegistry = _RedisPubsubRegistry(_redis_conn)
"""Redis Pubsub registry singleton."""


class RedisMixin(object):
    event = 'update'

    def __init__(self, *args, **kwargs):
        super(RedisMixin, self).__init__(*args, **kwargs)
        RedisPubsubRegistry.register(self.get_channel(), self)

    def cleanup(self):
        RedisPubsubRegistry.unregister(self.get_channel(), self)
        super(RedisMixin, self).cleanup()

    @classmethod
    def get_channel(cls):
        return cls.channel

    @classmethod
    def publish(cls, channel=None, **data):
        channel = channel or cls.get_channel()
        RedisPubsubRegistry.publish(channel, serialize(data))

    def check_and_send(self, msg):
        self.send(self.event, **deserialize(msg['data']))


class RedisRoomsMixin(object):

    def __init__(self, *args, **kwargs):
        self.rooms = set()
        super(RedisRoomsMixin, self).__init__(*args, **kwargs)

    @classmethod
    def _get_room_name(cls, room):
        return '.'.join([cls.channel, str(room)])

    @classmethod
    def get_channel(cls):
        return super(RedisRoomsMixin, cls).get_channel() + '.*'

    @classmethod
    def publish(cls, room, channel=None, **data):
        if channel is None:
            channel = cls._get_room_name(room)
        super(RedisRoomsMixin, cls).publish(channel=channel, **data)

    def join(self, room):
        self.rooms.add(self._get_room_name(room))

    def leave(self, room):
        self.rooms.remove(self._get_room_name(room))

    def on_join(self, room):
        self.join(room)

    def on_leave(self, room):
        self.leave(room)

    def check_and_send(self, msg):
        if msg['channel'] in self.rooms:
            super(RedisRoomsMixin, self).check_and_send(msg)



from django.contrib.gis.geos import Point, LinearRing, Polygon
from django.contrib.gis.geoip import GeoIP
from django.conf import settings
import os
import math


__all__ = ['geoip', 'bearing', 'GeoVector', 'GeoCoord', 'GeoPolygon']


import logging
log = logging.getLogger('gravy.geo')


geoip = GeoIP()


# geoip2
try:
    from geoip2.database import Reader
    from geoip2.errors import AddressNotFoundError

    __all__.extend(['geoip2', 'AddressNotFoundError'])

    # TODO: how much work is it to make this GeoIP compatible?
    class GeoIP2(object):
        path = '/usr/share/GeoIP'
        filenames = {
            'city': 'GeoIP2-City.mmdb',
            'country': 'GeoIP2-Country.mmdb',
            'isp': 'GeoIP2-ISP.mmdb',
        }

        def __init__(self, raw=True):
            self.raw = raw
            self.readers = {}

        def __getattr__(self, name):
            if name not in self.readers:
                if name not in self.filenames:
                    return super(GeoIP2, self).__getattr__(name)
                filename = os.path.join(
                    getattr(settings, 'GEOIP_PATH', self.path),
                    getattr(settings, 'GEOIP2_' + name.upper(), self.filenames[name])
                )
                self.readers[name] = Reader(filename)
            return getattr(self.readers[name], name)

    geoip2 = GeoIP2()
except ImportError:
    pass


EARTH_RADIUS = 6378137
COS = [math.cos(math.radians(a)) for a in range(360)]
SIN = [math.sin(math.radians(a)) for a in range(360)]


def _degrees(azimuth):
    return (810 - azimuth) % 360;


def bearing(lat1, lon1, lat2, lon2):
    rlat1 = math.radians(lat1)
    rlat2 = math.radians(lat2)
    rlon1 = math.radians(lon1)
    rlon2 = math.radians(lon2)
    dlon = math.radians(lon2-lon1) 
    b = math.atan2(
        math.sin(dlon) * math.cos(rlat2),
        math.cos(rlat1) * math.sin(rlat2) - \
        math.sin(rlat1) * math.cos(rlat2) * math.cos(dlon)
    )
    bd = math.degrees(b)
    bn = (bd+360)%360
    return bn


class GeoVector(object):
    """
    GeoVector in cartesian space.

    Typical used to represent a distance in metres from the origin when used
    for geographical operations. Can be thought of as the flat earth cartesian
    coordinates, as in, East 2m and North 5m.
    """
    def __init__(self, east, north):
        self.east = east
        self.north = north

    @classmethod
    def OnCircle(cls, radius, angle):
        a = angle % 360
        return cls(radius * COS[a], radius * SIN[a])

    @classmethod
    def OnEllipse(cls, semimajor, semiminor, alpha, beta):
        a = alpha % 360
        b = beta % 360
        return cls(
            semimajor * COS[a] * COS[b] + semiminor * SIN[a] * SIN[b],
            semiminor * SIN[a] * COS[b] - semimajor * COS[a] * SIN[b]
        )


class GeoCoord(object):
    """
    Coordinate on the earth.
    """
    def __init__(self, latitude, longitude):
        self.latitude = latitude
        self.longitude = longitude

    def __str__(self):
        return '(%f, %f)' % (self.latitude, self.longitude)

    def __eq__(self, other):
        return self.latitude == other.latitude and self.longitude == other.longitude

    def __ne__(self, other):
        return not (self == other)

    @classmethod
    def FromGeosPoint(cls, point):
        return cls(point.y, point.x)

    def ToGeosPoint(self):
        return Point(self.longitude, self.latitude)

    def travel(self, vector):
        """
        Get the destination coordinate that results in traveling along the
        given vector from this coordinate.
        """
        latitude = self.latitude + math.degrees(math.asin(vector.north / EARTH_RADIUS))
        longitude = self.longitude + math.degrees(math.asin(vector.east / math.cos(math.radians(latitude)) / EARTH_RADIUS))
        return self.__class__(latitude, longitude)

    def bearing(self, other):
        return bearing(self.latitude, self.longitude, other.latitude, other.longitude)


class GeoPolygon(object):
    """
    Polygon on the earth.
    """
    def __init__(self, *vertices):
        self.vertices = vertices

    def __iter__(self):
        return iter(self.vertices)

    @classmethod
    def FromCircle(cls, origin, radius, step=10):
        vertices = []
        for d in range(0, 360, step):
            vertices.append(origin.travel(GeoVector.OnCircle(radius, _degrees(d))))
        vertices.append(vertices[0])
        return cls(*vertices)

    @classmethod
    def FromEllipse(cls, origin, semimajor, semiminor, angle, step=10):
        vertices = []
        for d in range(360, 0, -step):
            vertices.append(origin.travel(GeoVector.OnEllipse(semimajor, semiminor, _degrees(d), angle)))
        vertices.append(vertices[0])
        return cls(*vertices)

    @classmethod
    def FromArc(cls, origin, inradius, outradius, startangle, stopangle, step=10):
        vertices = []
        align = stopangle + step
        if inradius:
            for d in range(startangle, stopangle, step):
                vertices.append(origin.travel(GeoVector.OnCircle(inradius, _degrees(d))))
            align = d
            vertices.append(origin.travel(GeoVector.OnCircle(inradius, _degrees(stopangle))))
        else:
            vertices.append(origin)
        vertices.append(origin.travel(GeoVector.OnCircle(outradius, _degrees(stopangle))))
        for d in range(align, startangle, -step):
            vertices.append(origin.travel(GeoVector.OnCircle(outradius, _degrees(d))))
        vertices.append(origin.travel(GeoVector.OnCircle(outradius, _degrees(startangle))))
        vertices.append(vertices[0])        
        return cls(*vertices)

    def ToGeosPolygon(self):
        return Polygon(LinearRing([v.ToGeosPoint() for v in self.vertices]))

    def _split_into_chunks(self, value):
        while value >= 32: #2^5, while there are at least 5 bits
            # first & with 2^5-1, zeros out all the bits other than the first five
            # then OR with 0x20 if another bit chunk follows
            yield (value & 31) | 0x20 
            value >>= 5
        yield value

    def _encode_value(self, value):
        # Step 2 & 4
        value = ~(value << 1) if value < 0 else (value << 1)
        # Step 5 - 8
        chunks = self._split_into_chunks(value)
        # Step 9-10
        return (chr(chunk + 63) for chunk in chunks)

    def encode(self):
        """
        Encodes a polyline using Google"s polyline algorithm
        
        See http://code.google.com/apis/maps/documentation/polylinealgorithm.html 
        for more information.
        """
        result = []
        prev_lat = 0
        prev_lng = 0
        for vertex in self:        
            lat, lng = int(vertex.latitude * 1e5), int(vertex.longitude * 1e5)
            d_lat = self._encode_value(lat - prev_lat)
            d_lng = self._encode_value(lng - prev_lng)
            prev_lat, prev_lng = lat, lng
            result.append(d_lat)
            result.append(d_lng)
        return ''.join(c for r in result for c in r)

    def centroid(self):
        result = GeoCoord(0.0, 0.0)
        signedArea = 0.0
        v0 = self.vertices[0]
        for i, v1 in enumerate(self.vertices[1:]):
            a = v0.latitude * v1.longitude - v0.longitude * v1.latitude
            signedArea += a
            result.latitude += (v0.latitude + v1.latitude) * a
            result.longitude += (v0.longitude + v1.longitude) * a
            v0 = v1
        signedArea *= 0.5
        result.latitude /= (6.0 * signedArea)
        result.longitude /= (6.0 * signedArea)
        return result



from django.core.cache import cache as _cache


class CachedProperty(property):
    """
    Decorator much like django cached_property however it also caches to a
    'real' cache and exposes some additional functionality for setting/deleting
    the cache value along with the ability to perform additional actions when
    the cache value is set or deleted
    """
    # use the default django cache
    cache = _cache

    # use str on the object (this should be overridden)
    key_fmt = '{object}'

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        key = self.key_fmt.format(object=obj)
        value = self.cache.get(key)
        if value is not None:
            return value
        value = super(CachedProperty, self).__get__(obj, objtype)
        self.cache.set(key, value)
        return value

    def __set__(self, obj, value):
        key = self.key_fmt.format(object=obj)
        self.cache.set(key, value)
        if self.fset is not None:
            super(CachedProperty, self).__set__(obj, value)

    def __delete__(self, obj):
        key = self.key_fmt.format(object=obj)
        self.cache.delete(key)
        if self.fdel is not None:
            super(CachedProperty, self).__delete__(obj)

    on_set = property.setter

    on_del = property.deleter


class CachedClassProperty(CachedProperty):
    """
    Same as CachedProperty decorator but acts as a class proprty rather than an
    instance property. 
    """
    def __get__(self, obj, objtype=None):
        if objtype is None:
            objtype = type(obj)
        return super(CachedClassProperty, self).__get__(objtype)

    def __set__(self, obj):
        super(CachedClassProperty, self).__set__(type(obj))

    def __delete__(self, obj):
        super(CachedClassProperty, self).__delete__(type(obj))


def real_cached_property(key_fmt, cache=None):
    """
    Return a CachedProperty decorator that utilises the suppiled key_fmt and
    cache.
    """
    class RealCachedProperty(CachedProperty):
        def __init__(self, *args, **kwargs):
            super(RealCachedProperty, self).__init__(*args, **kwargs)
            self.key_fmt = key_fmt
            if cache is not None:
                self.cache = cache

    return RealCachedProperty


def real_cached_classproperty(key_fmt, cache=None):
    """
    Return a CachedClassProperty decorator that utilises the suppiled key_fmt
    and cache.
    """
    class RealCachedClassProperty(CachedClassProperty):
        def __init__(self, *args, **kwargs):
            super(RealCachedClassProperty, self).__init__(*args, **kwargs)
            self.key_fmt = key_fmt
            if cache is not None:
                self.cache = cache

    return RealCachedClassProperty




from django.utils.module_loading import autodiscover_modules


import logging
log = logging.getLogger('gravy.websocket')


__all__ = ['autodiscover',]


def autodiscover():
    autodiscover_modules('sockets')



from ctypes.util import find_library
from ctypes import *


__all__ = [
    'liblzo', 'LZO_E_OK', 'LZO_E_ERROR', 'LZO_E_OUT_OF_MEMORY',
    'LZO_E_NOT_COMPRESSIBLE', 'LZO_E_INPUT_OVERRUN', 'LZO_E_OUTPUT_OVERRUN',
    'LZO_E_LOOKBEHIND_OVERRUN', 'LZO_E_EOF_NOT_FOUND',
    'LZO_E_INPUT_NOT_CONSUMED', 'LZO_E_NOT_YET_IMPLEMENTED',
    'LZO_E_INVALID_ARGUMENT', 'LZO_E_INVALID_ALIGNMENT',
    'LZO_E_OUTPUT_NOT_CONSUMED', 'LZO_E_INTERNAL_ERROR', 'lzo_buffer',
    'lzo1x_1_compress', 'lzo1x_decompress'
]


liblzo = CDLL(find_library('lzo2'))


# types
lzo_uint = c_uint
lzo_uintp = POINTER(lzo_uint)
lzo_bytep = POINTER(c_ubyte)
lzo_voidp = c_void_p
lzo_sizeof_dict_t = sizeof(lzo_bytep)


# constants
LZO1_MEM_COMPRESS = 8192 * lzo_sizeof_dict_t
LZO1X_MEM_COMPRESS = LZO1X_1_MEM_COMPRESS = 16384 * lzo_sizeof_dict_t
LZO1_99_MEM_COMPRESS = 65536 * lzo_sizeof_dict_t
LZO1_MEM_DECOMPRESS = 0


# errors
LZO_E_OK = 0
LZO_E_ERROR = -1
LZO_E_OUT_OF_MEMORY = -2
LZO_E_NOT_COMPRESSIBLE = -3
LZO_E_INPUT_OVERRUN = -4
LZO_E_OUTPUT_OVERRUN = -5
LZO_E_LOOKBEHIND_OVERRUN = -6
LZO_E_EOF_NOT_FOUND = -7
LZO_E_INPUT_NOT_CONSUMED = -8
LZO_E_NOT_YET_IMPLEMENTED = -9
LZO_E_INVALID_ARGUMENT = -10
LZO_E_INVALID_ALIGNMENT = -11
LZO_E_OUTPUT_NOT_CONSUMED = -12
LZO_E_INTERNAL_ERROR = -99


# lzo_buffer
def lzo_buffer(size):
    return create_string_buffer(size)


# lzo1x_1_compress
def lzo1x_1_compress(src, dst=None):
    if dst is None:
        dst = lzo_buffer(len(src) + len(src)/16 + 64 + 3)
    dst_len = lzo_uint(sizeof(dst))
    wrkmem = create_string_buffer(LZO1X_1_MEM_COMPRESS)
    result = liblzo.lzo1x_1_compress(src, len(src), dst, byref(dst_len), wrkmem)
    if result != LZO_E_OK:
        raise Exception("LZO Error: %d" % result)
    return dst[:dst_len.value]


# lzo1x_decompress
def lzo1x_decompress(src, dst):
    dst_len = lzo_uint(sizeof(dst))
    result = liblzo.lzo1x_decompress(src, len(src), dst, byref(dst_len), None)
    if result != LZO_E_OK:
        raise Exception("LZO Error: %d" % result)
    return dst[:dst_len.value]


# test
if __name__ == '__main__':
    import sys
    plain = sys.stdin.read()
    compressed = lzo1_compress(plain)
    decompressed = lzo1_decompress(compressed, lzo_buffer(len(plain)))
    sys.stdout.write(decompressed)



from django.forms.fields import *
from django.forms.models import ModelMultipleChoiceField
from django.core.files.storage import DefaultStorage
from django.core.validators import MinValueValidator, MaxValueValidator
from django.core.exceptions import ValidationError
from django.utils.html import format_html
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from ..utils import epoch
from .widgets import *
from betterforms.forms import BetterForm, Fieldset
from collections import OrderedDict
import jsonschema
import json
import copy
import six
import re


import logging
log = logging.getLogger('gravy.forms.fields')


__all__ = [
    'NamedMultiValueField', 'RepeatNamedMultiValueField', 'SavedFileField',
    'SerializedDateTimeField', 'SerializedModelMultipleChoiceField',
    'SeparatedFieldMixin', 'SeparatedChoiceField', 'SeparatedCharField',
    'FilePermissionsField', 'MultipleFileField', 'SchemaParseError',
    'validate_json', 'FormGenerator', 'normalise_schema', 'form_from_schema',
    'FieldsetMixin', 'FieldsetFormGenerator', 'fieldsetform_from_schema',
    'fieldsetform_from_schemas', 'FormSchemaField', 'FieldsetFormSchemaField',
]

# proxy django.forms.fields
import django.forms.fields
__all__.extend(django.forms.fields.__all__)
# proxy .widgets
import widgets
__all__.extend(widgets.__all__)


# TODO: the whole labels, help_texts, initial thing could be changed a bit
class NamedMultiValueField(MultiValueField):
    widget = NamedMultiWidget
    fields = ()

    def __init__(self, widget=None, fields=None, *args, **kwargs):
        kwargs.setdefault('require_all_fields', False)
        self.named_fields = OrderedDict(fields or self.fields)
        widgets, initial, help_texts, labels = OrderedDict(), {}, {}, {}
        for name, field in six.iteritems(self.named_fields):
            widgets[name] = field.widget
            if hasattr(field, 'initial'):
                initial[name] = field.initial
            if hasattr(field, 'label'):
                labels[name] = field.label
            if hasattr(field, 'help_text'):
                help_texts[name] = field.help_text
        widget = self.widget(widgets=widgets, labels=labels, help_texts=help_texts, initial=initial)
        fields = self.named_fields.values()
        super(NamedMultiValueField, self).__init__(
            widget=widget, fields=fields, initial=initial,
            *args, **kwargs
        )

    def compress(self, data_list):
        return dict(zip(self.named_fields.keys(), data_list))


# TODO: add ability to set intial number of repetitions
class RepeatNamedMultiValueField(NamedMultiValueField):
    widget = RepeatNamedMultiWidget

    def clean(self, value):
        if not isinstance(value, (list, tuple)):
            raise ValidationError(self.error_messages['invalid'], code='invalid')
        return [super(RepeatNamedMultiValueField, self).clean(v) for v in value]


class SavedFileField(FileField):
    """
    A FileField that saves the file using the supplied storage. The field value
    is the filename of the saved file.
    """
    storage_class = DefaultStorage

    def get_storage(self):
        return self.storage_class()

    def save(self, value):
        return self.get_storage().save(value.name, value)

    def to_python(self, value):
        value = super(SavedFileField, self).to_python(value)
        if value in self.empty_values:
            return None
        return self.save(value)


class SerializedDateTimeField(DateTimeField):
    """
    A DateTimeField that serializes the returned date.
    """
    widget = SerializedDateTimeInput

    def to_python(self, value):
        value = super(SerializedDateTimeField, self).to_python(value)
        if value in self.empty_values:
            return None
        return epoch(value)


class SerializedModelMultipleChoiceField(ModelMultipleChoiceField):

    def clean(self, value):
        qs = super(SerializedModelMultipleChoiceField, self).clean(value)
        return [object.pk for object in qs]


class FilePermissionsField(CharField):
    default_validators = [
        MinValueValidator(0),
        MaxValueValidator(0777),
    ]

    default_error_messages = {
        'invalid': _('Enter a whole number.') 
    }

    def to_python(self, value):
        value = super(FilePermissionsField, self).to_python(value)
        if value in self.empty_values:
            return None
        try:
            value = int(value, 0)
        except (ValueError, TypeError):
            raise ValidationError(self.error_messages['invalid'], code='invalid')
        return value


class SeparatedFieldMixin(object):
    def to_python(self, value):
        if isinstance(value, list):
            return [super(SeparatedFieldMixin, self).to_python(v) for v in value]
        return super(SeparatedFieldMixin, self).to_python(value)


class SeparatedChoiceField(SeparatedFieldMixin, ChoiceField):
    widget = SeparatedSelect


# TODO: add validation that it is a list etc
class SeparatedCharField(SeparatedFieldMixin, CharField):
    widget = SeparatedTextInput


class MultipleFileField(FileField):
    widget = MultipleFileInput

    def to_python(self, value):
        return [super(MultipleFileField, self).to_python(v) for v in value]


class SchemaParseError(ValueError):
    def __init__(self, message, context=None):
        super(SchemaParseError, self).__init__(message)
        self.context = context

    @property
    def schema(self):
        if self.context is None:
            return ''
        if isinstance(self.context, six.string_types):
            return self.context
        return json.dumps(self.context, indent=4)

    @property
    def html(self):
        return format_html('{}<br><pre>{}</pre>', self.message, self.schema)


def validate_json(obj, schema, **kwargs):
    try:
        jsonschema.validate(obj, schema)
    except jsonschema.ValidationError as e:
        message = kwargs.get('message', e.message)
        context = kwargs.get('context', obj)
        raise SchemaParseError(message, context)


class FormGenerator(object):
    """
    Parses a form fields schema to actual form fields. When parsing issues are
    encountered a SchemaParseError is raised (when subclassing this methodology
    should continue to be used).
    """
    def __init__(self, fields, validate=True):
        if validate:
            self.validate(fields)
        self.__fields = fields

    def get_form_class(self):
        self.formfields = OrderedDict()
        for field in self.__fields:
            self.formfields[field['name']] = self.create_field(field)
        return type('DynamicForm', (BetterForm,), self.formfields)

    @classmethod
    def validate(cls, fields):
        validate_json(fields, {'type': 'array'})
        for field in fields:
            cls.validate_field(field)

    @staticmethod
    def get_field_options(field):
        options = {}
        options['label'] = field.get('label', '')
        options['help_text'] = field.get('help_text')
        options['required'] = field.get('required', False)
        if 'initial' in field:
            options['initial'] = field['initial']
        if field.get('hidden', False):
            options['widget'] = HiddenInput
        return options

    @classmethod
    def validate_field_options(cls, field):
        schema = {
            'type': 'object',
            'properties': {
                'label': {'type': 'string'},
                'help_text': {'type': 'string'},
                'required': {'type': 'boolean'},
                'hidden': {'type': 'boolean'}
            }
        }
        validate_json(field, schema)
        return cls.get_field_options(field)

    def create_field(self, field):
        options = self.get_field_options(field)
        return getattr(self, 'create_field_for_' + field['type'])(field, options)

    @classmethod
    def validate_field(cls, field):
        options = cls.validate_field_options(field)
        schema = {
            'type': 'object',
            'properties': {
                'type': {'type': 'string'}
            }, 'required': ['type']
        }
        validate_json(field, schema)
        if not hasattr(cls, 'create_field_for_' + field['type']):
            raise SchemaParseError('The field type "%s" is not supported.' % field['type'], field)
        func = getattr(cls, 'validate_field_for_' + field['type'], None)
        if func is not None:
            func(field, options)

    @staticmethod
    def create_field_for_text(field, options):
        options['max_length'] = field.get('max_length', 20)
        return CharField(**options)

    @staticmethod
    def validate_field_for_text(field, options):
        schema = {
            'type': 'object',
            'properties': {
                'type': {'enum': ['text']},
                'initial': {'type': 'string'},
                'max_length': {'type': 'integer', 'minimum': 0}
            }, 'required': ['type']
        }
        validate_json(field, schema)

    @staticmethod
    def create_field_for_file(field, options):
        options['max_length'] = field.get('max_length', 512)
        return SavedFileField(**options)

    @staticmethod
    def validate_field_for_file(field, options):
        schema = {
            'type': 'object',
            'properties': {
                'type': {'enum': ['file']},
                'initial': {'type': 'string'},
                'max_length': {'type': 'integer', 'minimum': 0}
            }, 'required': ['type']
        }
        validate_json(field, schema)

    @staticmethod
    def create_field_for_textarea(field, options):
        options['max_length'] = field.get('max_length', 9999)
        return CharField(widget=Textarea(attrs={'rows': '5'}), **options)

    @staticmethod
    def validate_field_for_textarea(field, options):
        schema = {
            'type': 'object',
            'properties': {
                'type': {'enum': ['textarea']},
                'initial': {'type': 'string'},
                'max_length': {'type': 'integer', 'minimum': 0}
            }, 'required': ['type']
        }
        validate_json(field, schema)

    @staticmethod
    def create_field_for_integer(field, options):
        options['min_value'] = field.get('min_value', -999999999)
        options['max_value'] = field.get('max_value', 999999999)
        return IntegerField(**options)

    @staticmethod
    def validate_field_for_integer(field, options):
        schema = {
            'type': 'object',
            'properties': {
                'type': {'enum': ['integer']},
                'initial': {'type': 'integer'},
                'min_value': {'type': 'integer'},
                'max_value': {'type': 'integer'}
            }, 'required': ['type']
        }
        validate_json(field, schema)
        if field.get('min_value', -999999999) > field.get('max_value', 999999999):
            raise SchemaParseError('min_value must be less than or equal to max_value.', field)

    @staticmethod
    def create_field_for_datetime(field, options):
        return DateTimeField(**options)

    @staticmethod
    def validate_field_for_datetime(field, options):
        schema = {
            'type': 'object',
            'properties': {
                'type': {'enum': ['datetime']},
                'initial': {'type': 'string'}
            }, 'required': ['type']
        }
        validate_json(field, schema)

    @staticmethod
    def create_field_for_perm(field, options):
        return FilePermissionsField(**options)

    @staticmethod
    def validate_field_for_perm(field, options):
        schema = {
            'type': 'object',
            'properties': {
                'type': {'enum': ['perm']},
                'initial': {'type': 'integer'}
            }, 'required': ['type']
        }
        validate_json(field, schema)

    @staticmethod
    def create_field_for_radio(field, options):
        options['choices'] = [(c['value'], c['name']) for c in field['choices']]
        return ChoiceField(widget=RadioSelect, **options)

    @staticmethod
    def validate_field_for_radio(field, options):
        schema = {
            'type': 'object',
            'properties': {
                'type': {'enum': ['radio']},
                'choices': {
                    'type': 'array',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'name': {'type': 'string'},
                            'value': {},
                        }, 'required': ['name', 'value']
                    }
                }
            }, 'required': ['type', 'choices']
        }
        validate_json(field, schema)
        sentinel = object()
        initial = field.get('initial', sentinel)
        if initial is not sentinel and initial not in [c['value'] for c in field['choices']]:
            raise SchemaParseError('initial must match one of the values given in choices.', field)

    @staticmethod
    def create_field_for_select(field, options):
        options['choices'] = [(c['value'], c['name']) for c in field['choices']]
        return ChoiceField(**options)

    @staticmethod
    def validate_field_for_select(field, options):
        schema = {
            'type': 'object',
            'properties': {
                'type': {'enum': ['select']},
                'choices': {
                    'type': 'array',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'name': {'type': 'string'},
                            'value': {},
                        }, 'required': ['name', 'value']
                    }
                }
            }, 'required': ['type', 'choices']
        }
        validate_json(field, schema)
        sentinel = object()
        initial = field.get('initial', sentinel)
        if initial is not sentinel and initial not in [c['value'] for c in field['choices']]:
            raise SchemaParseError('initial must match one of the values given in choices.', field)

    @staticmethod
    def create_field_for_selectlist(field, options):
        options['choices'] = [(c['value'], c['name']) for c in field['choices']]
        return SeparatedChoiceField(**options)

    @staticmethod
    def validate_field_for_selectlist(field, options):
        schema = {
            'type': 'object',
            'properties': {
                'type': {'enum': ['selectlist']},
                'choices': {
                    'type': 'array',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'name': {'type': 'string'},
                            'value': {'type': 'array'},
                        }, 'required': ['name', 'value']
                    }
                }
            }, 'required': ['type', 'choices']
        }
        validate_json(field, schema)
        sentinel = object()
        initial = field.get('initial', sentinel)
        if initial is not sentinel and initial not in [c['value'] for c in field['choices']]:
            raise SchemaParseError('initial must match one of the values given in choices.', field)

    @staticmethod
    def create_field_for_checkbox(field, options):
        return BooleanField(**options)

    @staticmethod
    def validate_field_for_checkbox(field, options):
        schema = {
            'type': 'object',
            'properties': {
                'type': {'enum': ['checkbox']},
                'initial': {'type': 'boolean'},
            }, 'required': ['type']
        }
        validate_json(field, schema)

    def create_field_for_multiple(self, field, options):
        options['fields'] = []
        for subfield in field['fields']:
            options['fields'].append((subfield['name'], self.create_field(subfield)))
        return NamedMultiValueField(**options)

    @classmethod
    def validate_field_for_multiple(cls, field, options):
        schema = {
            'type': 'object',
            'properties': {
                'type': {'enum': ['multiple']},
                'fields': {'type': 'array'}
            }, 'required': ['type', 'fields']
        }
        validate_json(field, schema)
        for subfield in field['fields']:
            cls.validate_field(subfield)

    def create_field_for_repeat(self, field, options):
        options['fields'] = []
        for subfield in field['fields']:
            options['fields'].append((subfield['name'], self.create_field(subfield)))
        return RepeatNamedMultiValueField(**options)

    @classmethod
    def validate_field_for_repeat(cls, field, options):
        schema = {
            'type': 'object',
            'properties': {
                'type': {'enum': ['repeat']},
                'fields': {'type': 'array'}
            }, 'required': ['type', 'fields']
        }
        validate_json(field, schema)
        for subfield in field['fields']:
            cls.validate_field(subfield)


def normalise_schema(schema):
    if hasattr(schema, 'read'):
        schema = schema.read()
    if isinstance(schema, six.string_types):
        try:
            normalised_schema = json.loads(schema)
        except ValueError as e:
            raise SchemaParseError(e, schema)
        return normalised_schema
    # since the schema may be modified we need a copy
    return copy.deepcopy(schema)


def form_from_schema(schema, generator_class=FormGenerator, validate=True):
    """
    Accepts a form schema in string, fileobject or list format, normalises it
    and returns a form class that reflects the suppiled schema.

    Raises a ValueError with the appropriate message the schema is invalid.
    """
    normalised_schema = normalise_schema(schema)
    generator = generator_class(normalised_schema, validate=validate)
    return generator.get_form_class()


class FieldsetMixin(object):
    """
    Mixin to tranparently handle the namespacing that is added by the fieldset
    form generator. It adds an additional dictionary level to cleaned_data and
    reformats namespacing for intial. See FieldsetFormGenerator.create_field()
    """
    def __init__(self, *args, **kwargs):
        if 'initial' in kwargs and kwargs['initial']:
            initial = {}
            for fieldset_name, fieldset in six.iteritems(kwargs['initial']):
                for field_name, value in six.iteritems(fieldset):
                    name = '.'.join([fieldset_name, field_name])
                    initial[name] = value
            kwargs['initial'] = initial
        super(FieldsetMixin, self).__init__(*args, **kwargs)

    @cached_property
    def cleaned_fieldset_data(self):
        cleaned = {}
        for name, value in six.iteritems(self.cleaned_data):
            fieldset_name, field_name = name.split('.', 1)
            if not fieldset_name in cleaned:
                cleaned[fieldset_name] = {}
            cleaned[fieldset_name][field_name] = value
        return cleaned


class FieldsetFormGenerator(FormGenerator):
    """
    Parses a fieldset form fields schema to actual form fields and formsets.
    When parsing issues are encountered a SchemaParseError is raised (when
    subclassing this methodology should continue to be used).
    """
    def __init__(self, fieldsets, validate=True):
        if validate:
            self.validate(fieldsets)
        self.__fieldsets = fieldsets

    def get_form_class(self):
        self.formfields = OrderedDict()
        self.fieldsets = []
        for fieldset in self.__fieldsets:
            self.fieldsets.append(self.create_fieldset(fieldset))
        class Meta:
            fieldsets = tuple(self.fieldsets)
        self.formfields['Meta'] = Meta
        return type('DynamicFieldsetForm', (FieldsetMixin, BetterForm,), self.formfields)

    @classmethod
    def validate(cls, fieldsets):
        validate_json(fieldsets, {'type': 'array'})
        for fieldset in fieldsets:
            cls.validate_fieldset(fieldset)

    @staticmethod
    def get_fieldset_options(fieldset):
        options = {}
        options['legend'] = fieldset['legend']
        options['help_text'] = fieldset.get('help_text')
        options['is_hidden'] = fieldset.get('hidden', False)
        return options

    @staticmethod
    def validate_fieldset_options(fieldset):
        schema = {
            'type': 'object',
            'properties': {
                'legend': {'type': 'string'},
                'help_text': {'type': 'string'},
                'hidden': {'type': 'boolean'}
            }, 'required': ['legend']
        }
        validate_json(fieldset, schema)

    def create_fieldset(self, fieldset):
        options = self.get_fieldset_options(fieldset)
        options['fields'] = []
        for field in fieldset['fields']:
            # namespace field name with fieldset name
            field['name'] = '.'.join([fieldset['name'], field['name']])
            f = self.create_field(field)
            self.formfields[field['name']] = f
            options['fields'].append(field['name'])
        options['fields'] = tuple(options['fields'])
        fs = Fieldset(fieldset['name'], **options)
        return fs

    @classmethod
    def validate_fieldset(cls, fieldset):
        cls.validate_fieldset_options(fieldset)
        schema = {
            'type': 'object',
            'properties': {
                'name': {'type': 'string'},
                'fields': {'type': 'array'}
            }, 'required': ['name', 'fields']
        }
        validate_json(fieldset, schema)
        for field in fieldset['fields']:
            cls.validate_field(field)


def _merge_fields(a, b):
    a.extend(b)
    return a


def _merge_fieldset(a, b):
    for k, v in six.iteritems(b):
        if k == 'fields':
            _merge_fields(a[k], b[k])
        else:
            a[k] = v
    return a


def _merge_schema(a, b):
    e = []
    for bf in b:
        bd = False
        for i, af in enumerate(a):
            if af['name'] == bf['name']:
                _merge_fieldset(af, bf)
                bd = True
        if not bd:
            e.append(bf)
    a.extend(e)
    return a


def fieldsetform_from_schema(schema, generator_class=FieldsetFormGenerator, validate=True):
    """
    Accepts a fieldset form schema in string, fileobject or list format,
    normalises it and returns a form class that reflects the suppiled schema.

    Raises a ValueError with the appropriate message the schema is invalid.
    """
    return form_from_schema(schema, generator_class=generator_class, validate=validate)


def fieldsetform_from_schemas(*schemas, **kwargs):
    """
    Accepts one or more fieldset form schemas in string, fileobject or list
    format, normalises and merges them into a single schema and returns a form
    class that reflects the merged schema.

    The order of the fieldsets in the merged schema is first in best dressed
    i.e. fieldsets in the first schema will be first in the merged schema and
    so on.

    The legend and description of a fieldset will be overwritten by each schema
    that defines a fieldset with the same name (later schemas overriding
    earlier schemas).

    Fields can be added to a fieldset (or overwritten) by each schema that
    defines a fieldset with the same name (later schemas overriding earlier
    schemas).

    Raises a ValueError with the appropriate message any of the schemas are
    invalid.
    """
    generator_class = kwargs.get('generator_class', FieldsetFormGenerator)
    normalised_schemas = map(normalise_schema, schemas)
    # validate the schemas before merge
    for normalised_schema in normalised_schemas:
        generator_class.validate(normalised_schema)
    merged = reduce(_merge_schema, normalised_schemas, [])
    return fieldsetform_from_schema(merged, generator_class=generator_class, validate=False)


class FormSchemaField(CharField):
    """
    A field for entering a form schema and having it validated and returned as
    a django form.
    """
    widget = Textarea
    default_error_messages = {
        'invalid': _('Not a valid schema: %(schema)s')
    }
    generator_class = FormGenerator

    def __init__(self, *args, **kwargs):
        generator_class = kwargs.pop('generator_class', None)
        if generator_class is not None:
            self.generator_class = generator_class
        super(FormSchemaField, self).__init__(*args, **kwargs)

    def get_form(self, value):
        return form_from_schema(value, generator_class=self.generator_class)

    def to_python(self, value):
        value = super(FormSchemaField, self).to_python(value)
        if value in self.empty_values:
            return None
        try:
            return self.get_form(value)
        except SchemaParseError as e:
            raise ValidationError(self.error_messages['invalid'], code='invalid', params={'schema': e.schema})


class FieldsetFormSchemaField(FormSchemaField):
    """
    A field for entering a fieldset form schema and having it validated and
    returned as a django form.
    """
    generator_class = FieldsetFormGenerator

    def get_form(self, value):
        return fieldsetform_from_schema(value,
            generator_class=self.generator_class
        )



import json


import logging
log = logging.getLogger('gravy.websocket.helpers')


__all__ = ['serialize', 'deserialize',]


class _AlwaysEncode(json.JSONEncoder):
    def default(self, obj):
        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError:
            pass
        return unicode(obj)


def serialize(obj):
    return json.dumps(obj, cls=_AlwaysEncode)


def deserialize(data):
    return json.loads(data)



from django.utils import six
from PIL import Image
from PIL.ExifTags import TAGS
from pytesseract import image_to_string


__all__ = ['dump', 'load', 'ocr', 'exif', 'reorient', 'thumbnail']


import logging
log = logging.getLogger('gravy.images')


def dump(img, fd, **kwargs):
    img.save(fd, **kwargs)


def load(fd):
    return Image.open(fd)


def ocr(img):
    return image_to_string(
        img.resize((6*d for d in img.size), Image.NEAREST)
    )


def exif(img):
    data = {}
    info = img._getexif()
    if info is None:
        return data
    for tag, value in six.iteritems(info):
        name = TAGS.get(tag, tag)
        if name in ('UserComment',):
            encoding = value[:8]
            value = value[8:]
        data[name] = value
    return data


def reorient(img):
    rotate = exif(img).get('Orientation', 1)
    if rotate == 3:
        return img.rotate(-180)
    elif rotate == 6:
        return img.rotate(-90)
    elif rotate == 8:
        return img.rotate(-270)
    return img


def thumbnail(img, width=300, height=300):
    i = img.copy()
    i.thumbnail((width, height), Image.ANTIALIAS)
    return i



from geventwebsocket.gunicorn.workers import GeventWebSocketWorker
from .handler import DjangoWebSocketHandler


class DjangoWebSocketWorker(GeventWebSocketWorker):
    wsgi_handler = DjangoWebSocketHandler



import os
from setuptools import setup, find_packages


with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


setup(
    name='django-gravy',
    version='0.1.6',
    packages=find_packages(),

    # only install the most common requirements
    install_requires = [
        'django-betterforms',
        'python-magic',
        'pyfifo'
    ],
    include_package_data=True,
    license='BSD License',
    description='Reusable extras for Django (the gravy).',
    long_description=README,
    author='greenbender',
    author_email='byron.platt@gmail.com',
    url='https://github.com/greenbender/django-gravy',
    zip_safe=False,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



from geventwebsocket import WebSocketError
from .redis import RedisMixin, RedisRoomsMixin
from .helpers import serialize, deserialize
import gevent


import logging
log = logging.getLogger('gravy.websocket.namespace')


__all__ = [
    'NamespaceHandlerRegistry', 'BaseNamespace', 'JsonNamespace',
    'RedisNamespace', 'RedisRoomsNameSpace'
]


class _NamespaceHandlerRegistry(object):
    """
    Registry for keeping track of namespace handlers.
    """

    def __init__(self):
        self._registry = {}

    def register(self, channel, handler):
        self._registry[channel] = handler

    def __getitem__(self, channel):
        return self._registry[channel]

    def get(self, channel):
        return self[channel]


NamespaceHandlerRegistry = _NamespaceHandlerRegistry()
"""Namespace registry singleton."""


class namespace(object):
    """
    Decorator for Namespace handlers.
    """
    registry = NamespaceHandlerRegistry

    def __init__(self, channel=''):
        self.channel = channel
 
    def __call__(self, handler):
        handler.channel = self.channel
        self.registry.register(self.channel, handler)
        return handler


class BaseNamespace(object):
    """
    Base Namespace handler.
    """
    socket_name = 'wsgi.websocket'
    channel = None

    def __init__(self, request):
        self.request = request
        self.environ = self.get_environ()
        self.socket = self.get_socket()
        self.running = False

    def get_environ(self):
        return self.request.environ

    def get_socket(self):
        return self.environ[self.socket_name]
    
    def recv(self):
        return self.socket.receive()
    
    def send(self, pkt):
        self.socket.send(pkt)
 
    def setup(self):
        pass

    def cleanup(self):
        pass

    def run(self):
        self.running = True
        self.setup()
        try:
            while self.running:
                self.recv()
        except WebSocketError as e:
            log.debug(e)
        except Exception as e:
            log.error(e)
        finally:
            self.stop()
            self.cleanup()

    def stop(self):
        self.running = False

class JsonNamespace(BaseNamespace):
    """
    JsonNamespace handler.
    """
    ping_timeout = 30
    
    def recv(self):
        pkt = super(JsonNamespace, self).recv()
        if pkt is None:
            return
        msg = deserialize(pkt)
        if 'event' not in msg:
            log.error('Invalid Websocket packet')
            return
        method = getattr(self, 'on_' + msg['event'], None)
        if method is None:
            return
        data = msg.get('data', None)
        return method(data)
    
    def send(self, event, **data):
        msg = {'event': event}
        if data:
            msg['data'] = data
        pkt = serialize(msg)
        super(JsonNamespace, self).send(pkt)
 
    def ping_run(self):
        try:
            while not self.ping_event.wait(self.ping_timeout):
                self.send('ping')
        except WebSocketError as e:
            log.debug(e)
        except Exception as e:
            log.error(e)
        finally:
            self.stop()

    def setup(self):
        super(JsonNamespace, self).setup()
        if self.ping_timeout is not None:
            self.ping_event = gevent.event.Event()
            self.ping_task = gevent.spawn(self.ping_run)
        
    def cleanup(self):
        if self.ping_timeout is not None:
            self.ping_event.set()
            self.ping_task.join()
        super(JsonNamespace, self).cleanup()


class RedisNamespace(RedisMixin, JsonNamespace):
    pass


class RedisRoomsNamespace(RedisRoomsMixin, RedisMixin, JsonNamespace):
    pass



from django import template
from django.template.defaultfilters import pluralize
from ..utils import seconds_to_units


register = template.Library()


@register.filter
def humanize_seconds(value):
    qty, unit = seconds_to_units(abs(value))
    unit += pluralize(qty) 
    if value < 0:
        qty = -qty 
    return "{0} {1}".format(qty, unit)



#!/usr/bin/python

from struct import pack, unpack, calcsize
from fifo import Fifo
from _binding import *


__all__ = [
    'LZO_MAGIC', 'MAX_BLOCK', 'Compress', 'compress', 'Decompress', 'decompress'
]
import _binding
__all__.extend(_binding.__all__)


LZO_MAGIC = 'LZO\x00'
MAX_BLOCK = 0x7fff

_BLOCK_HEADER_FORMAT = 'H'
_BLOCK_HEADER_LENGTH = calcsize(_BLOCK_HEADER_FORMAT)


class Compress(object):

    def __init__(self):
        self.__fifo = Fifo()
        self.__output = []
        self.__output.append(LZO_MAGIC)

    def __yield_output(self):
        output = ''.join(self.__output)
        self.__output = []
        return output

    def compress(self, data):
        self.__fifo.write(data)

        while True:
            block = self.__fifo.read(MAX_BLOCK)
            if not block:
                break
            data = lzo1x_1_compress(block)
            if len(data) >= len(block):
                self.__output.append(pack(_BLOCK_HEADER_FORMAT, len(block)))
                self.__output.append(block)
            else:
                self.__output.append(pack(_BLOCK_HEADER_FORMAT, len(data) | 0x8000))
                self.__output.append(data)
        return self.__yield_output()
        

def compress(plain):
    return Compress().compress(plain)


class Decompress(object):

    STATE_MAGIC = 0
    STATE_BLOCK_HEADER = 1
    STATE_BLOCK = 2

    def __init__(self):
        self.__buff = lzo_buffer(MAX_BLOCK)
        self.__fifo = Fifo()
        self.__output = []
        self.state = self.STATE_MAGIC

    def __yield_output(self):
        output = ''.join(self.__output)
        self.__output = []
        return output

    def decompress(self, data):
        self.__fifo.write(data)

        while True:

            # magic
            if self.state == self.STATE_MAGIC:
                magic = self.__fifo.readblock(len(LZO_MAGIC))
                if not magic:
                    break
                if magic != LZO_MAGIC:
                    raise Exception('Bad magic: %s' % magic)
                self.state = self.STATE_BLOCK_HEADER

            # block header
            if self.state == self.STATE_BLOCK_HEADER:
                header = self.__fifo.readblock(_BLOCK_HEADER_LENGTH)
                if not header:
                    break
                self.__block_length = unpack(_BLOCK_HEADER_FORMAT, header)[0]
                self.__block_compressed = self.__block_length & 0x8000
                self.__block_length ^= self.__block_compressed
                if self__block_length:
                    self.state = self.STATE_BLOCK

            # block
            if self.state == self.STATE_BLOCK:
                block = self.__fifo.readblock(self.__block_length)
                if not block:
                    break
                if self.__block_compressed:
                    decomp = lzo1x_decompress(block, self.__buff)
                    self.__output.append(decomp)
                else:
                    self.__output.append(block)
                self.state = self.STATE_BLOCK_HEADER

        return self.__yield_output()


def decompress(data):
    return Decompress().decompress(data)


if __name__ == '__main__':
    import os
    import sys
    name = os.path.basename(sys.argv[0])
    if name == 'compress':
        l = Compress()
        while True:
            data = sys.stdin.read(4096)
            if not data:
                break
            sys.stdout.write(l.compress(data))
    elif name == 'decompress':
        l = Decompress()
        while True:
            data = sys.stdin.read(1024)
            if not data:
                break
            sys.stdout.write(l.decompress(data))



from geventwebsocket.handler import WebSocketHandler


class DjangoWebSocketHandler(WebSocketHandler):

    def run_websocket(self):
        """
        Just run the websocket request on the application. Dont use client
        tracking because we are using unix domain sockets and the client addr
        is the same for every socket.
        """
        try:
            self.application(self.environ, lambda s, h, e=None: [])
        finally:
            # not sure all this tidy up is necessary
            self.environ.update({'wsgi.websocket': None})
            if not self.websocket.closed:
                self.websocket.close()
            self.websocket = None

    @property
    def active_client(self):
        """
        Since we are tracking clients return None here.
        """
        return



from django.db.models import signals
from django.contrib.auth.middleware import RemoteUserMiddleware
from .db.models.registration import CurrentUserFieldRegistry


import logging
log = logging.getLogger('gravy.middleware')


__all__ = [
    'DebugMiddleware', 'NginxRemoteUserMiddleware', 'CurrentUserMiddleware'
]


class DebugMiddleware(object):

    def process_request(self, request):
        log.error(request.META)


class NginxRemoteUserMiddleware(RemoteUserMiddleware):
    header = 'HTTP_REMOTE_USER'


class CurrentUserMiddleware(object):

    def process_request(self, request):
        if request.method in ('GET', 'HEAD', 'OPTION', 'TRACE'):
            return
        if hasattr(request, 'user') and request.user.is_authenticated():
            user = request.user
        else:
            user = None
        def update_user_fields(sender, instance, **kwargs):
            registry = CurrentUserFieldRegistry()
            if sender in registry:
                for field in registry.get_fields(sender):
                    # XXX: this isn't a great check for add but i'm not sure
                    # of a better way at this point
                    add = instance.pk is None and getattr(instance, field.name) is None
                    if field.auto_user or (field.auto_user_add and add):
                        setattr(instance, field.name, user)
        signals.pre_save.connect(update_user_fields, dispatch_uid=request, weak=False)

    def process_response(self, request, response):
        signals.pre_save.disconnect(dispatch_uid=request)
        return response



from django.db import models
from django.conf import settings

from .registration import CurrentUserFieldRegistry


class CurrentUserField(models.ForeignKey):

    def __init__(self, auto_user=False, auto_user_add=False, **kwargs):
        kwargs['to'] = settings.AUTH_USER_MODEL
        kwargs.setdefault('null', True)
        kwargs.setdefault('editable', False)
        self.auto_user = auto_user
        self.auto_user_add = auto_user_add
        super(CurrentUserField, self).__init__(**kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(CurrentUserField, self).deconstruct()
        # XXX: should be able to del 'to' keyword but there is a bug
        # Check https://code.djangoproject.com/ticket/24434#ticket for a
        #       resolution every now and then
        #del kwargs['to']
        if kwargs['null'] is True:
            del kwargs['null']
        if kwargs['editable'] is False:
            del kwargs['editable']
        if self.auto_user:
            kwargs['auto_user'] = True
        if self.auto_user_add:
            kwargs['auto_user_add'] = True
        return name, path, args, kwargs

    def contribute_to_class(self, cls, name):
        super(CurrentUserField, self).contribute_to_class(cls, name)
        if not cls._meta.abstract:
            registry = CurrentUserFieldRegistry()
            registry.add_field(cls, self)

